#include <malloc.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <stdlib.h>

#include "util.h"
#include "object.h"
#include "classmember.h"
#include "class.h"
#include "directory.h"
#include "source.h"
#include "collection.h"
#include "file.h"
#include "function.h"

#include "demangle.h"

Cfunction		*Cfunction::g_hash[HASHSIZE] = {0};
unsigned int	Cfunction::g_hash_cnt        = 0;

/*
static void trap_function(void)
{
}
*/

Cfunction::Cfunction(const char *nameP) : Cclassmember(nameP)
{
	m_hashP       = 0;
}

void
Cfunction::reset(void)
{
	int 		i;
	Cfunction	*functionP;

	for (i = 0; i < HASHSIZE; ++i) {
		for (functionP = g_hash[i]; functionP; functionP = functionP->m_hashP) {
			functionP->m_flags |= InvisibleX;
}	}	}

Cfunction *
Cfunction::local_exists(const char *nameP, Centity *parentP)
{
	unsigned int	index;
	Cfunction		*functionP;

	index = hash(nameP);
	for (functionP = g_hash[index]; functionP; functionP = functionP->m_hashP) {
		if (!strcmp(functionP->m_nameP, nameP)) {
			if (functionP->m_flags & InvisibleX) {
				continue;
			}
			if (functionP->m_flags & LocalX) {
				if (functionP->m_parentP == parentP) {
					return functionP;
	}	}	}	}
	return(0);
}

Cfunction *
Cfunction::global_exists(const char *nameP)
{
	unsigned int	index;
	Cfunction		*functionP;

	index = hash(nameP);
	for (functionP = g_hash[index]; functionP; functionP = functionP->m_hashP) {
		if (!strcmp(functionP->m_nameP, nameP)) {
			if (functionP->m_flags & InvisibleX) {
				continue;
			}
			if (!(functionP->m_flags & LocalX)) {
				return functionP;
	}	}	}
	return(0);
}

Cfunction *
Cfunction::locate(const char *nameP, int flags, Csource *sourceP)
{
	unsigned int	index;
	Cfunction		**functionPP, *functionP, *bestP;
	Csource			*parentP;

/*
	if (!strcmp(nameP, "get_value")) {
		trap_function();
	}
*/
	bestP         = 0;
	index         = hash(nameP);
	for (functionPP = g_hash + index; (functionP = *functionPP); functionPP = &(functionP->m_hashP)) {
		if (!strcmp(functionP->m_nameP, nameP)) {
			if (functionP->m_flags & InvisibleX) {
				continue;
			}
			parentP = (Csource *) functionP->m_parentP;
			if ((flags & LocalX) || (functionP->m_flags & LocalX)) {
				if (sourceP != parentP) {
					continue;
			}	}
			if (!bestP) {
				bestP = functionP;
			} else if (parentP == sourceP) {
				bestP = functionP;
			}
	}	}
	if (bestP) {
		if (!(flags & FindX)) {
			if (!bestP->m_parentP) {
				/* This function was called from outside our source
				 * by something which didn't know where to find it
				 */
				if (flags & LocalX) {
					/* Can't be a function called outside our source */
					goto create_new;
				}
				/* Presume we are the location of the called function */

				bestP->m_parentP = sourceP;
				bestP->add_created();
			} else if (bestP->m_parentP != sourceP) {
				goto create_new;
		}	}
		return bestP;
	}

create_new:
	functionP = new Cfunction(nameP);
	if (!functionP) {
		outofmemory();
	}
	/* Add to end so always return first function declaration seen if
	 * searching for location of the function declaration:
	 *
     * eg. if we are looking for foo() and foo() was in foo.c later
     *     also seen in foo.o return reference to foo.c:foo() in
	 *     preferance to foo.o:foo()
     */
	if (!(flags & FindX) && sourceP) {
		functionP->m_parentP = sourceP;
		functionP->add_created();
	}
	*functionPP = functionP;
	++g_hash_cnt;
	return(functionP);
}

void
Cfunction::resolve(void)
{
	Csource 	*sourceP;
	Cfunction	*functionP;
	int			i;

	sourceP = 0;
	for (i = 0; i < HASHSIZE; ++i) {
		for (functionP = g_hash[i]; functionP; functionP = functionP->m_hashP) {
			if (!functionP->m_parentP) {
				if (!sourceP) {
					sourceP    = new Ccollection(":library:", 0);
					if (!sourceP) {
						outofmemory();
					}
				}
				functionP->m_parentP = sourceP;
				functionP->m_flags  |= ExternalX;
				functionP->add_created();
}	}	}	}

/* virtual */ objectE
Cfunction::type(void) const
{
	return(functionE);
}

/* virtual */ void
Cfunction::emit_id(void) const
{
	fprintf(stdout, "F%p", this);
}

const char *
Cfunction::get_type(void) const
{
	int			flags;
	const char 	*typeP;

	flags = m_flags;
	if (flags & LocalX) {
		typeP = "LF";
	} else if (flags & ExternalX) {
		typeP = "EF";
	} else {
		typeP = "GF";
	}
	return(typeP);
}

/* virtual */ void
Cfunction::emit_edges(void)
{
	Cclassmember::emit_edges();
	if (Cclass::g_headP) {
		emit_class();
	}
}
		
/* virtual */ void
Cfunction::emit_instance(void) const
{
	fprintf(stdout, "$INSTANCE F%p %s\n", this, get_type());
}

void
Cfunction::emit_attributes(void) const
{
	const char 	*nameP;
	char		*demangleP;

	emit_id();
	fputs(" {", stdout);

	nameP     = m_nameP;
	demangleP = 0;
	if (nameP[0] == '_') {
		demangleP = cplus_demangle(nameP, DMGL_ANSI|DMGL_PARAMS|DMGL_GNU_V3);
		if (demangleP) {
			fprintf(stdout,"mangle=\"%s\" ", nameP);
			nameP = demangleP;
	}	}
	fprintf(stdout, "label=\"%s\"", nameP);
	if (demangleP) {
		free(demangleP);
	}
	if (m_fileP) {
		char	*filenameP = m_fileP->m_nameP;
		char 	path_to_me[2048];

		path_to_me[0] = 0;
		full_path(path_to_me);

		// TODO: Improve this logic using ../ etc.
		if (!strcmp(filenameP, path_to_me)) {
			filenameP = ".";
		}
		fprintf(stdout, " file=\"%s\" lineno=%d", filenameP, m_lineno);
	}
	fputs("}\n", stdout);
}

/* virtual */ void
Cfunction::print(void) const
{
	fputs(m_nameP, stdout);
}
